import argparse
from pathlib import Path

from .clean import CleanConfig, clean_nb_file
from .core import get_nb_names

parser = argparse.ArgumentParser(
    prog="nbclean",
    description="Clean metadata and execution_count from Jupyter notebooks.",
)
parser.add_argument(
    "path",
    default=".",
    nargs="*",
    help="Path for nb or folder with notebooks.",
)
parser.add_argument(
    "-s",
    "--silent",
    action="store_true",
    help="Silent mode.",
)
parser.add_argument(
    "--not-pt",
    action="store_true",
    help="Do not preserve timestamp.",
)


def app() -> None:
    """Clean metadata and execution_count from Jupyter notebook."""
    cfg = parser.parse_args()
    path_list = cfg.path if isinstance(cfg.path, list) else [cfg.path]
    nb_files: list[Path] = []
    if not cfg.silent:
        print(f"Path: {', '.join(cfg.path)}, preserve timestamp: {not cfg.not_pt}")
    for path in path_list:
        try:
            nb_files.extend(get_nb_names(path))
        except FileNotFoundError:
            print(f"{path} not exists!")
    if not cfg.silent:
        print(f"notebooks to check: {len(nb_files)} ")
    cleaned, errors = clean_nb_file(
        nb_files,
        CleanConfig(
            silent=cfg.silent,
            preserve_timestamp=not cfg.not_pt,
        ),
    )
    if not cfg.silent:
        print(f"cleaned nbs: {len(cleaned)}")
        if errors:
            print(f"with errors: {len(errors)}")
            for nb, exc in errors:
                print(f"{nb}: {exc}")


if __name__ == "__main__":
    app()



from pathlib import Path

from nbmetaclean.core import get_nb_names, read_nb, write_nb


def test_read_nb():
    """test read notebook"""
    file = Path("tests/test_nbs/test_nb_1.ipynb")
    nb = read_nb(file)
    assert isinstance(nb, dict)
    assert nb["metadata"]["language_info"] == {"name": "python"}
    assert nb["metadata"]["authors"][0]["name"] == "Andrei Yasyrev"
    assert nb["nbformat"] == 4
    assert nb["nbformat_minor"] == 2
    cells = nb["cells"]
    assert isinstance(cells, list)
    assert len(cells) == 2
    # markdown
    assert cells[0]["cell_type"] == "markdown"
    assert cells[0]["source"] == []
    assert cells[0]["metadata"] == {}
    # code
    assert cells[1]["cell_type"] == "code"
    assert cells[1]["source"] == []
    assert cells[1]["execution_count"] is None
    assert cells[1]["metadata"] == {}
    assert cells[1]["outputs"] == []


def test_write_nb(tmp_path: Path):
    """test write notebook"""
    file = Path("tests/test_nbs/test_nb_1.ipynb")
    nb = read_nb(file)
    write_nb(nb, tmp_path / file.name)
    with open(tmp_path / file.name, "r", encoding="utf-8") as fh:
        res_text = fh.read()
    with open(file, "r", encoding="utf-8") as fh:
        org_text = fh.read()
    assert res_text == org_text

    # write with name w/o suffix
    result = write_nb(nb, tmp_path / "test_nb_1")
    assert result == tmp_path / "test_nb_1.ipynb"


def test_get_nb_names():
    """test get_nb_names"""
    path = Path("tests/test_nbs")
    # filename as argument
    file = path / "test_nb_1.ipynb"
    names = get_nb_names(file)
    assert len(names) == 1
    names.sort(key=lambda x: x.name)
    assert names[0] == file
    # path as argument
    names = get_nb_names(path)
    assert len(names) == 2
    names.sort(key=lambda x: x.name)
    assert names[0] == file
    # path as argument. add hidden files
    names = get_nb_names(path, hidden=True)
    assert len(names) == 3
    try:
        get_nb_names("wrong_name")
        assert False
    except FileNotFoundError as ex:
        assert True
        assert str(ex) == "wrong_name not exists!"


def test_get_nb_names_recursive_hidden(tmp_path: Path):
    """test get_nb_names recursive hidden"""
    suffix = ".ipynb"
    # add one nb
    with open((tmp_path / "tst").with_suffix(suffix), "w", encoding="utf-8") as _:
        pass
    files = get_nb_names(tmp_path)
    assert len(files) == 1

    # add hidden nb
    with open((tmp_path / ".tst").with_suffix(suffix), "w", encoding="utf-8") as _:
        pass
    files = get_nb_names(tmp_path)
    assert len(files) == 1
    files = get_nb_names(tmp_path, hidden=True)
    assert len(files) == 2
    # add simple file
    with open((tmp_path / "simple"), "w", encoding="utf-8") as _:
        pass
    files = get_nb_names(tmp_path)
    assert len(files) == 1

    # add dir with one nb, hidden nb
    new_dir = tmp_path / "new_dir"
    new_dir.mkdir()
    with open((new_dir / "tst").with_suffix(suffix), "w", encoding="utf-8") as _:
        pass
    with open((new_dir / ".tst").with_suffix(suffix), "w", encoding="utf-8") as _:
        pass
    files = get_nb_names(tmp_path)
    assert len(files) == 2
    files = get_nb_names(tmp_path, hidden=True)
    assert len(files) == 4

    files = get_nb_names(tmp_path, recursive=False)
    assert len(files) == 1

    # add hidden dir
    hid_dir = tmp_path / ".hid_dir"
    hid_dir.mkdir()
    with open((hid_dir / "tst").with_suffix(suffix), "w", encoding="utf-8") as _:
        pass
    with open((hid_dir / ".tst").with_suffix(suffix), "w", encoding="utf-8") as _:
        pass
    files = get_nb_names(tmp_path, hidden=True)
    assert len(files) == 6
    files = get_nb_names(tmp_path)
    assert len(files) == 2



import copy
import os
from pathlib import Path

from pytest import CaptureFixture

from nbmetaclean.clean import (
    NB_METADATA_PRESERVE_MASKS,
    CleanConfig,
    clean_cell,
    clean_nb,
    clean_nb_file,
    filter_meta_mask,
    filter_metadata,
)
from nbmetaclean.core import read_nb, write_nb


def test_get_meta_by_mask():
    """test get_meta_by_mask"""
    nb = read_nb(Path("tests/test_nbs/.test_nb_2_meta.ipynb"))
    nb_meta = nb.get("metadata")

    # string as nb_meta
    new_meta = filter_meta_mask("some string")
    assert new_meta == "some string"

    # no mask
    new_meta = filter_meta_mask(nb_meta)
    assert new_meta == {}

    # mask
    nb_meta["some key"] = "some value"
    new_meta = filter_meta_mask(nb_meta, ("some key",))
    assert new_meta == {"some key": "some value"}
    new_meta = filter_meta_mask(nb_meta, NB_METADATA_PRESERVE_MASKS[0])
    assert new_meta == {"language_info": {"name": "python"}}

    # mask for empty result
    new_meta = filter_meta_mask(nb_meta, ("some other key",))
    assert new_meta == {}


def test_new_metadata():
    """test new_metadata"""
    nb_meta = read_nb("tests/test_nbs/.test_nb_2_meta.ipynb").get("metadata")
    new_meta = filter_metadata(nb_meta)
    assert isinstance(new_meta, dict)
    assert not new_meta
    new_meta = filter_metadata(nb_meta, [("language_info", "name")])
    assert new_meta == {"language_info": {"name": "python"}}


def test_clean_nb_metadata():
    """test clean_nb_metadata"""
    test_nb = read_nb("tests/test_nbs/test_nb_2_clean.ipynb")
    cfg = CleanConfig()
    result = clean_nb(test_nb, cfg)
    assert not result

    # add metadata, new filter, mask not merged
    test_nb["metadata"]["some key"] = "some value"
    cfg.nb_metadata_preserve_mask = (("some key",),)
    cfg.mask_merge = False
    result = clean_nb(test_nb, cfg)
    assert result
    assert test_nb["metadata"] == {"some key": "some value"}

    # add metadata, new filter, mask merged
    test_nb = read_nb("tests/test_nbs/test_nb_2_clean.ipynb")
    test_nb["metadata"]["some_key"] = {"key_1": 1, "key_2": 2}
    cfg.nb_metadata_preserve_mask = (("some_key", "key_1"),)
    cfg.mask_merge = True
    result = clean_nb(test_nb, cfg)
    assert result
    assert test_nb["metadata"]["authors"][0]["name"] == "Andrei Yasyrev"
    assert test_nb["metadata"]["some_key"] == {"key_1": 1}


def test_clean_cell_metadata():
    """test clean_cell_metadata"""
    test_nb = read_nb("tests/test_nbs/.test_nb_2_meta.ipynb")

    # clear outputs
    cell = copy.deepcopy(test_nb.get("cells")[1])
    assert cell.get("outputs")
    assert not cell.get("metadata")
    assert cell.get("execution_count") == 1
    cell["metadata"] = {"some key": "some value"}
    changed = clean_cell(
        cell,
        cfg=CleanConfig(
            clear_outputs=True,
            clear_cell_metadata=True,
        ),
    )
    assert changed
    assert not cell.get("outputs")
    assert not cell.get("metadata")
    assert not cell.get("execution_count")

    # dont clear outputs, execution_count, mask
    cell = copy.deepcopy(test_nb.get("cells")[1])
    cell["metadata"] = {"some key": "some value"}
    cell["outputs"][0]["metadata"] = {
        "some key": "some value",
        "some other key": "some value",
    }
    changed = clean_cell(
        cell,
        CleanConfig(
            clear_execution_count=False,
            clear_cell_metadata=True,
            cell_metadata_preserve_mask=(("some key",),),
        ),
    )
    assert changed
    assert cell["outputs"][0]["metadata"] == {"some key": "some value"}
    assert cell["metadata"] == {"some key": "some value"}
    assert cell["execution_count"] == 1

    # clear outputs, same mask -> no changes meta, clear execution_count
    changed = clean_cell(
        cell,
        cfg=CleanConfig(),
    )
    assert changed
    assert cell["execution_count"] is None
    assert cell["metadata"] == {"some key": "some value"}

    # clear execution_count, metadata
    changed = clean_cell(
        cell,
        cfg=CleanConfig(
            clear_cell_metadata=True,
        ),
    )
    assert changed
    assert not cell["outputs"][0]["metadata"]
    assert not cell["execution_count"]
    assert not cell["metadata"]
    assert not cell["outputs"][0]["metadata"]


def test_clean_cell():
    """test clean_cel"""
    test_nb = read_nb("tests/test_nbs/.test_nb_2_meta.ipynb")

    # nothing to clean.
    cell = copy.deepcopy(test_nb.get("cells")[1])
    assert cell.get("outputs")
    assert not cell.get("metadata")
    assert cell.get("execution_count") == 1
    result = clean_cell(cell, CleanConfig(clear_execution_count=False))
    assert not result

    # clean cell metadata, cell without metadata
    cell["metadata"] = {}
    result = clean_cell(cell, CleanConfig(clear_cell_metadata=True))
    assert result
    assert not cell.get("metadata")
    assert cell.get("outputs")

    # clear output metadata
    cell["outputs"][0]["metadata"] = {"some key": "some value"}
    result = clean_cell(
        cell,
        CleanConfig(
            clear_cell_metadata=True,
            cell_metadata_preserve_mask=(("some key",),),
        ),
    )
    assert not result
    assert cell["outputs"][0].get("metadata") == {"some key": "some value"}


def test_clean_cell_metadata_markdown():
    """test clean_cell_metadata with markdown cell"""
    test_nb = read_nb("tests/test_nbs/.test_nb_2_meta.ipynb")
    cell = copy.deepcopy(test_nb["cells"][0])
    cell["metadata"] = {"some key": "some value"}
    changed = clean_cell(
        cell,
        cfg=CleanConfig(
            clear_cell_metadata=True,
        ),
    )
    assert changed
    assert not cell["metadata"]


def test_clean_nb():
    """test clean nb"""
    path = Path("tests/test_nbs")
    nb_path = path / ".test_nb_2_meta.ipynb"
    nb_clean = path / "test_nb_2_clean.ipynb"
    nb = read_nb(nb_path)
    assert nb["cells"][1]["execution_count"] == 1
    assert nb["cells"][1]["outputs"][0]["execution_count"] == 1
    assert nb["metadata"]
    result = clean_nb(nb, cfg=CleanConfig())
    assert result is True
    assert nb["cells"][1]["execution_count"] is None
    assert nb["cells"][1]["outputs"][0]["execution_count"] is None
    nb_clean = read_nb(nb_clean)
    assert nb == nb_clean

    # # try clean cleaned
    result = clean_nb(nb_clean, cfg=CleanConfig())
    assert not result

    # # clean metadata, leave execution_count
    nb = read_nb(nb_path)
    result = clean_nb(
        nb,
        cfg=CleanConfig(clear_execution_count=False),
    )
    assert result
    assert nb["cells"][1]["execution_count"] == 1
    assert nb["cells"][1]["outputs"][0]["execution_count"] == 1
    assert nb["metadata"] == nb_clean["metadata"]

    # clean nb metadata, leave cells metadata
    nb = read_nb(nb_path)
    nb["cells"][1]["metadata"] = {"some key": "some value"}
    result = clean_nb(nb, CleanConfig(clear_execution_count=False))
    assert result
    assert nb["metadata"] == nb_clean["metadata"]
    assert nb["cells"][1]["metadata"] == {"some key": "some value"}
    assert nb["cells"][1]["execution_count"] == 1

    # clean cells metadata, leave nb metadata
    nb = read_nb(nb_path)
    nb_meta = copy.deepcopy(nb["metadata"])
    result = clean_nb(nb, CleanConfig(clear_nb_metadata=False))
    assert result
    assert nb["metadata"] == nb_meta
    assert nb["cells"][1]["execution_count"] is None


def test_clean_nb_file(tmp_path: Path, capsys: CaptureFixture[str]):
    """test clean nb file"""
    path = Path("tests/test_nbs")
    nb_name = ".test_nb_2_meta.ipynb"
    nb_clean = read_nb(path / "test_nb_2_clean.ipynb")

    # prepare temp test notebook
    test_nb_path = write_nb(read_nb(path / nb_name), tmp_path / nb_name)

    # clean meta, leave execution_count
    cleaned, errors = clean_nb_file(
        test_nb_path,
        cfg=CleanConfig(clear_execution_count=False),
    )
    assert len(cleaned) == 1
    assert len(errors) == 0
    nb = read_nb(cleaned[0])
    assert nb["metadata"] == nb_clean["metadata"]
    assert nb["cells"][1]["execution_count"] == 1
    assert nb["cells"][1]["outputs"][0]["execution_count"] == 1

    # clean meta, execution_count
    # path as list
    cleaned, errors = clean_nb_file([test_nb_path], CleanConfig())
    captured = capsys.readouterr()
    out = captured.out
    assert out.startswith("done")
    assert "test_clean_nb_file0/.test_nb_2_meta.ipynb" in out
    assert len(cleaned) == 1
    nb = read_nb(cleaned[0])
    assert nb == nb_clean

    # try clean cleaned
    cleaned, errors = clean_nb_file(test_nb_path, CleanConfig())
    assert len(cleaned) == 0
    captured = capsys.readouterr()
    out = captured.out
    assert not out.strip()

    # silent
    test_nb_path = write_nb(read_nb(path / nb_name), tmp_path / nb_name)
    cleaned, errors = clean_nb_file(test_nb_path, CleanConfig(silent=True))
    assert len(cleaned) == 1
    assert len(errors) == 0
    captured = capsys.readouterr()
    assert not captured.out.strip()


def test_clean_nb_file_errors(capsys: CaptureFixture[str], tmp_path: Path):
    """test clean_nb_file, errors"""
    path = tmp_path / "wrong_name"
    cleaned, errors = clean_nb_file(path)
    assert len(cleaned) == 0
    assert len(errors) == 1
    assert errors[0][0] == path
    assert "No such file or directory" in str(errors[0][1])
    captured = capsys.readouterr()
    assert not captured.out
    assert not captured.err
    with path.open("w", encoding="utf-8") as fh:
        fh.write("wrong nb")
    cleaned, errors = clean_nb_file(path)
    assert "wrong_name" in str(errors[0])
    assert len(cleaned) == 0
    assert len(errors) == 1
    captured = capsys.readouterr()
    assert not captured.out
    assert not captured.err


def test_clean_nb_file_timestamp(tmp_path: Path):
    """test clean_nb_file, timestamp"""
    path = Path("tests/test_nbs")
    nb_name = ".test_nb_2_meta.ipynb"
    nb_stat = (path / nb_name).stat()

    # prepare temp test notebook, set timestamp
    test_nb_path = write_nb(read_nb(path / nb_name), tmp_path / nb_name)
    os.utime(test_nb_path, (nb_stat.st_atime, nb_stat.st_mtime))
    test_nb_stat = test_nb_path.stat()
    assert test_nb_stat.st_atime == nb_stat.st_atime
    assert test_nb_stat.st_mtime == nb_stat.st_mtime

    cleaned, errors = clean_nb_file(test_nb_path)
    assert len(cleaned) == 1
    assert len(errors) == 0
    cleaned_stat = cleaned[0].stat()
    assert True
    assert cleaned_stat.st_mtime == test_nb_stat.st_mtime

    # dont preserve timestamp
    test_nb_path = write_nb(read_nb(path / nb_name), tmp_path / nb_name)
    os.utime(test_nb_path, (nb_stat.st_atime, nb_stat.st_mtime))
    cleaned, errors = clean_nb_file(test_nb_path, CleanConfig(preserve_timestamp=False))
    assert len(cleaned) == 1
    assert len(errors) == 0
    cleaned_stat = cleaned[0].stat()
    assert True
    assert cleaned_stat.st_mtime != nb_stat.st_mtime



from pathlib import Path, PosixPath
from typing import Dict, List, TypeVar, Union

PathOrStr = TypeVar("PathOrStr", Path, PosixPath, str)

NbNode = Dict[str, Union[str, int, "NbNode"]]
Metadata = Dict[str, Union[str, int, "Metadata"]]
Source = Union[str, List[str]]



from __future__ import annotations

import json
from pathlib import Path
from typing import Optional

from .typing import NbNode, PathOrStr


def read_nb(path: PathOrStr) -> NbNode:
    """Read notebook from filename.

    Args:
        path (Union[str, PosixPath): Notebook filename.

    Returns:
        Notebook: Jupyter Notebook as dict.
    """
    return json.load(open(path, "r", encoding="utf-8"))


def write_nb(
    nb: NbNode,
    path: PathOrStr,
) -> Path:
    """Write notebook to file

    Args:
        nb (Notebook): Notebook to write
        path (Union[str, PosixPath]): filename to write
    Returns:
        Path: Filename of writed Nb.
    """
    filename = Path(path)
    if filename.suffix != ".ipynb":
        filename = filename.with_suffix(".ipynb")
    with filename.open("w", encoding="utf-8") as fh:
        fh.write(
            json.dumps(
                nb,
                indent=1,
                separators=(",", ": "),
                ensure_ascii=False,
                sort_keys=True,
            )
            + "\n",
        )
    return filename


def get_nb_names(
    path: Optional[PathOrStr] = None,
    recursive: bool = True,
    hidden: bool = False,
) -> list[Path]:
    """Return list of notebooks from `path`. If no `path` return notebooks from current folder.

    Args:
        path (Union[Path, str, None]): Path for nb or folder with notebooks.
        recursive bool: Recursive search.
        hidden bool: Skip or not hidden paths, defaults to False.

    Raises:
        sys.exit: If filename or dir not exists or not nb file.

    Returns:
        List[Path]: List of notebooks names.
    """
    nb_path = Path(path or ".")

    if not nb_path.exists():
        raise FileNotFoundError(f"{nb_path} not exists!")

    if nb_path.is_dir():
        result = []
        for item in nb_path.iterdir():
            if item.is_file() and item.suffix == ".ipynb":
                if not hidden and item.name.startswith("."):
                    continue
                result.append(item)
            if item.is_dir():
                if recursive:
                    if not hidden and item.name.startswith("."):
                        continue
                    if "checkpoint" in item.name:
                        continue
                    result.extend(get_nb_names(item, recursive, hidden))

        return result

    return [nb_path]



from __future__ import annotations
import copy
from dataclasses import dataclass
import os

from pathlib import Path
from typing import Iterable, Optional, Union

from nbmetaclean.core import read_nb, write_nb

from .typing import NbNode, Metadata

NB_METADATA_PRESERVE_MASKS = (
    ("language_info", "name"),
    ("authors",),
)


@dataclass
class CleanConfig:
    """Clean config.

    Args:
        clear_nb_metadata (bool, optional): Clear notebook metadata. Defaults to True.
        clear_cell_metadata (bool, optional): Clear cell metadata. Defaults to False.
        clear_execution_count (bool, optional): Clear cell execution count. Defaults to True.
        clear_outputs (bool, optional): Clear cell outputs. Defaults to False.
        preserve_timestamp (bool, optional): Preserve timestamp. Defaults to True.
        silent (bool, optional): Silent mode. Defaults to False.
        nb_metadata_preserve_mask (Optional[tuple[str, ...]], optional):
            Preserve mask for notebook metadata. Defaults to None.
        cell_metadata_preserve_mask (Optional[tuple[str, ...]], optional):
            Preserve mask for cell metadata. Defaults to None.
        mask_merge (bool, optional): Merge masks. Add new mask to default.
            If False - use new mask. Defaults to True.
    """

    clear_nb_metadata: bool = True
    clear_cell_metadata: bool = False
    clear_execution_count: bool = True
    clear_outputs: bool = False
    preserve_timestamp: bool = True
    silent: bool = False
    nb_metadata_preserve_mask: Optional[Iterable[tuple[str, ...]]] = None
    cell_metadata_preserve_mask: Optional[Iterable[tuple[str, ...]]] = None
    mask_merge: bool = True


def filter_meta_mask(
    nb_meta: Union[str, int, Metadata],
    mask: Optional[Iterable[tuple[str, ...]]] = None,
) -> Union[str, int, Metadata]:
    """Filter metadata by mask. If no mask return empty dict."""
    if isinstance(nb_meta, (str, int)) or mask == ():
        return nb_meta
    if mask is None:
        return {}
    new_meta = {}
    value = nb_meta.get(mask[0])
    if value is not None:
        new_mask = tuple(mask[1:])
        new_meta[mask[0]] = filter_meta_mask(value, new_mask) or value
    return new_meta


def filter_metadata(
    nb_meta: Metadata,
    masks: Optional[list[tuple[str, ...]]] = None,
) -> Metadata:
    """Clean notebooknode metadata."""
    if masks is None:
        return {}
    filtered_meta: Metadata = {}
    for mask in masks:
        filtered_meta.update(filter_meta_mask(nb_meta, mask))  # type: ignore
    return filtered_meta


def clean_cell(
    cell: NbNode,
    cfg: CleanConfig,
) -> bool:
    """Clean cell: optionally metadata, execution_count and outputs."""
    changed = False

    if cfg.clear_cell_metadata:
        if metadata := cell.get("metadata", None):
            old_metadata = copy.deepcopy(metadata)
            cell["metadata"] = filter_metadata(
                metadata, cfg.cell_metadata_preserve_mask
            )
            if cell["metadata"] != old_metadata:
                changed = True

    if cfg.clear_execution_count and cell.get("execution_count"):
        cell["execution_count"] = None
        changed = True

    if cell.get("outputs"):
        if cfg.clear_outputs:
            cell["outputs"] = []
            changed = True
        elif cfg.clear_cell_metadata or cfg.clear_execution_count:
            result = clean_outputs(cell["outputs"], cfg)
            if result:
                changed = True

    return changed


def clean_outputs(outputs: list[NbNode], cfg: CleanConfig) -> bool:
    """Clean outputs."""
    changed = False
    for output in outputs:
        if cfg.clear_execution_count and output.get("execution_count", None):
            output["execution_count"] = None
            changed = True
        if cfg.clear_cell_metadata and (metadata := output.get("metadata", None)):
            old_metadata = copy.deepcopy(metadata)
            output["metadata"] = filter_metadata(
                metadata, cfg.cell_metadata_preserve_mask
            )
            if output["metadata"] != old_metadata:
                changed = True
    return changed


def clean_nb(
    nb: NbNode,
    cfg: CleanConfig,
) -> bool:
    """Clean notebook - metadata, execution_count, outputs.

    Args:
        nb (Notebook): Notebook to clean.
        clear_execution_count (bool, optional): Clear execution_count. Defaults to True.
        clear_outputs (bool, optional): Clear outputs. Defaults to False.

    Returns:
        bool: True if changed.
    """
    changed = False
    if cfg.clear_nb_metadata and (metadata := nb.get("metadata")):
        old_metadata = copy.deepcopy(metadata)
        masks = NB_METADATA_PRESERVE_MASKS
        if cfg.nb_metadata_preserve_mask:
            if not cfg.mask_merge:
                masks = cfg.nb_metadata_preserve_mask
            else:
                masks = cfg.nb_metadata_preserve_mask + masks

        nb["metadata"] = filter_metadata(metadata, masks=masks)
        if nb["metadata"] != old_metadata:
            changed = True
    if cfg.clear_cell_metadata or cfg.clear_execution_count or cfg.clear_outputs:
        for cell in nb["cells"]:
            result = clean_cell(
                cell,
                cfg,
            )
            if result:
                changed = True

    return changed


def clean_nb_file(
    path: Union[Path, list[Path]],
    cfg: Optional[CleanConfig] = None,
) -> tuple[list[Path], list[tuple[Path, Exception]]]:
    """Clean metadata and execution count from notebook.

    Args:
        path (Union[str, PosixPath]): Notebook filename or list of names.
        clear_nb_metadata (bool): Clear notebook metadata. Defaults to True.
        clear_cell_metadata (bool): Clear cell metadata. Defaults to False.
        clear_outputs (bool): Clear outputs. Defaults to False.
        preserve_timestamp (bool): Preserve timestamp. Defaults to True.
        clear_execution_count (bool, optional): Clean execution count. Defaults to True.
        silent (bool): Silent mode. Defaults to False.

    Returns:
        tuple[List[Path], List[TuplePath]]: List of cleaned notebooks, list of notebooks with errors.
    """
    if cfg is None:
        cfg = CleanConfig()
    if not isinstance(path, list):
        path = [path]
    cleaned: list[Path] = []
    errors: list[tuple[Path, Exception]] = []
    to_clean = len(path)
    for num, filename in enumerate(path):
        try:
            nb = read_nb(filename)
        except Exception as ex:
            errors.append((filename, ex))
            continue
        result = clean_nb(
            nb,
            cfg,
        )
        if result:
            cleaned.append(filename)
            if cfg.preserve_timestamp:
                stat = filename.stat()
            write_nb(nb, filename)
            if cfg.preserve_timestamp:
                os.utime(filename, (stat.st_atime, stat.st_mtime))
            if not cfg.silent:
                print(f"done {num + 1} of {to_clean}: {filename}")
    return cleaned, errors
